@@ -1,6 +1,6 @@
 ;;; haskell-decl-scan.el --- Declaration scanning module for Haskell Mode
 
-;; Copyright (C) 2004, 2005, 2007  Free Software Foundation, Inc.
+;; Copyright (C) 2004, 2005, 2007, 2009  Free Software Foundation, Inc.
 ;; Copyright (C) 1997-1998 Graeme E Moss
 
 ;; Author: 1997-1998 Graeme E Moss <gem@cs.york.ac.uk>
@@ -21,11 +21,8 @@
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs; see the file COPYING.  If not, write to the
-;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-;; Boston, MA 02111-1307, USA.
+;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-
 ;;; Commentary:
 
 ;; Purpose:
@@ -128,15 +125,36 @@
 ;;; Code:
 
 (require 'haskell-mode)
+(require 'syntax)
+(with-no-warnings (require 'cl))
 
-;;###autoload
+;;;###autoload
 ;; As `cl' defines macros that `imenu' uses, we must require them at
 ;; compile time.
 (eval-when-compile
-  (require 'cl)
   (condition-case nil
       (require 'imenu)
-    (error nil)))
+    (error nil))
+  ;; It makes a big difference if we don't copy the syntax table here,
+  ;; as Emacs 21 does, but Emacs 22 doesn't.
+  (unless (eq (syntax-table)
+	      (with-syntax-table (syntax-table) (syntax-table)))
+    (defmacro with-syntax-table (table &rest body)
+      "Evaluate BODY with syntax table of current buffer set to a copy of TABLE.
+The syntax table of the current buffer is saved, BODY is evaluated, and the
+saved table is restored, even in case of an abnormal exit.
+Value is what BODY returns."
+      (let ((old-table (make-symbol "table"))
+	    (old-buffer (make-symbol "buffer")))
+	`(let ((,old-table (syntax-table))
+	       (,old-buffer (current-buffer)))
+	   (unwind-protect
+	       (progn
+		 (set-syntax-table ,table)
+		 ,@body)
+	     (save-current-buffer
+	       (set-buffer ,old-buffer)
+	       (set-syntax-table ,old-table))))))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; General declaration scanning functions.
@@ -191,9 +209,9 @@ Point is not changed."
                     "\\(\\'\\)?\\s-*\\(\\s_+\\|`\\(\\sw+\\)`\\)")
                 "\\(\\sw+\\)?\\s-*\\(\\s_+\\|`\\(\\sw+\\)`\\)"))
              (let ((match2 (haskell-ds-match-string 2)))
-               ;; Weed out `::', `=' and `|' from potential infix
+               ;; Weed out `::', `∷',`=' and `|' from potential infix
                ;; symbolic variable.
-               (if (member match2 '("::" "=" "|"))
+               (if (member match2 '("::" "∷" "=" "|"))
                    ;; Variable identifier.
                    (haskell-ds-match-string 1)
                  (if (eq (aref match2 0) ?\`)
@@ -215,6 +233,16 @@ current line that starts with REGEXP and is not in `font-lock-comment-face'."
 		  (eq (get-text-property (point) 'face)
 		      'font-lock-comment-face)))))
 
+(defun haskell-ds-move-to-start-regexp-skipping-comments (inc regexp)
+  "Like haskell-ds-move-to-start-regexp, but uses syntax-ppss to
+  skip comments"
+  (let (p)
+    (loop
+     do (setq p (point))
+        (haskell-ds-move-to-start-regexp inc regexp)
+     while (and (nth 4 (syntax-ppss))
+                (/= p (point))))))
+
 (defvar literate-haskell-ds-line-prefix "> ?"
   "Regexp matching start of a line of Bird-style literate code.
 Current value is \"> \" as we assume top-level declarations start
@@ -260,7 +288,7 @@ then point does not move if already at the start of a declaration."
 	(bound (if direction (point-max) (point-min))))
     ;; Change syntax table.
     (with-syntax-table haskell-ds-syntax-table
-      ;; Move to beginning of line that starts the "current
+      ;; move to beginning of line that starts the "current
       ;; declaration" (dependent on DIRECTION and FIX), and then get
       ;; the variable typed or bound by this declaration, if any.
       (let ( ;; Where point was at call of function.
@@ -294,8 +322,8 @@ then point does not move if already at the start of a declaration."
           (if (and start (bobp))
               (setq abyss t)
             ;; Otherwise we move to the start of the first declaration
-            ;; on a line preceeding the current one.
-            (haskell-ds-move-to-start-regexp -1 start-decl-re))))
+            ;; on a line preceding the current one, skipping comments
+            (haskell-ds-move-to-start-regexp-skipping-comments -1 start-decl-re))))
       ;; If we are in the abyss, position and return as appropriate.
       (if abyss
           (if (not direction)
@@ -309,16 +337,17 @@ then point does not move if already at the start of a declaration."
             ;; declaration if moving backward, or move to the next
             ;; declaration if moving forward.
             (if direction
-                (haskell-ds-move-to-start-regexp 1 start-decl-re))
+                (haskell-ds-move-to-start-regexp-skipping-comments 1 start-decl-re))
           ;; If there is a variable, find the first
-          ;; succeeding/preceeding declaration that does not type or
-          ;; bind it.  Check for reaching start/end of buffer.
-          (haskell-ds-move-to-start-regexp increment start-decl-re)
+          ;; succeeding/preceding declaration that does not type or
+          ;; bind it.  Check for reaching start/end of buffer and
+          ;; comments.
+          (haskell-ds-move-to-start-regexp-skipping-comments increment start-decl-re)
           (while (and (/= (point) bound)
                       (and (setq newname (haskell-ds-get-variable line-prefix))
                            (string= name newname)))
             (setq name newname)
-            (haskell-ds-move-to-start-regexp increment start-decl-re))
+            (haskell-ds-move-to-start-regexp-skipping-comments increment start-decl-re))
           ;; If we are going backward, and have either reached a new
           ;; declaration or the beginning of a buffer that does not
           ;; start with a declaration, move forward to start of next
@@ -335,7 +364,7 @@ then point does not move if already at the start of a declaration."
                                            line-prefix))))
                        (and (not (looking-at start-decl-re))
                             (bobp))))
-              (haskell-ds-move-to-start-regexp 1 start-decl-re)))
+              (haskell-ds-move-to-start-regexp-skipping-comments 1 start-decl-re)))
         ;; Store whether we are at the start of a declaration or not.
         ;; Used to calculate final result.
         (let ((at-start-decl (looking-at start-decl-re)))
@@ -351,15 +380,15 @@ then point does not move if already at the start of a declaration."
   (and (boundp 'haskell-literate) (eq haskell-literate 'bird)))
 
 (defun haskell-ds-backward-decl ()
-  "Move point backward to the first character preceding the current
-point that starts a top-level declaration.  A series of declarations
-concerning one variable is treated as one declaration by this
-function.  So, if point is within a top-level declaration then move it
-to the start of that declaration.  If point is already at the start of
-a top-level declaration, then move it to the start of the preceding
-declaration.  Returns point if point is left at the start of a
-declaration, and nil otherwise, ie. because point is at the beginning
-of the buffer and no declaration starts there."
+  "Move backward to the first character that starts a top-level declaration.
+A series of declarations concerning one variable is treated as one
+declaration by this function.  So, if point is within a top-level
+declaration then move it to the start of that declaration.  If point
+is already at the start of a top-level declaration, then move it to
+the start of the preceding declaration.  Returns point if point is
+left at the start of a declaration, and nil otherwise, ie. because
+point is at the beginning of the buffer and no declaration starts
+there."
   (interactive)
   (haskell-ds-move-to-decl nil (haskell-ds-bird-p) nil))
 
@@ -395,8 +424,7 @@ positions and the type is one of the symbols \"variable\", \"datatype\",
       ;; Start and end of search space is currently just the starting
       ;; line of the declaration.
       (setq start (point)
-	    end   (progn (end-of-line) (point)))
-      (goto-char start)
+	    end   (line-end-position))
       (cond
        ;; If the start of the top-level declaration does not begin
        ;; with a starting keyword, then (if legal) must be a type
@@ -426,7 +454,7 @@ positions and the type is one of the symbols \"variable\", \"datatype\",
 	      (setq name-pos (match-beginning 1))
 	      (setq type 'class))))
        ;; Import declaration.
-       ((looking-at "import[ \t]+\\(qualified[ \t]+\\)?\\(\\sw+\\)")
+       ((looking-at "import[ \t]+\\(qualified[ \t]+\\)?\\(\\(?:\\sw\\|.\\)+\\)")
 	(setq name (haskell-ds-match-string 2))
 	(setq name-pos (match-beginning 2))
 	(setq type 'import))
@@ -471,6 +499,7 @@ positions and the type is one of the symbols \"variable\", \"datatype\",
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Declaration scanning via `imenu'.
 
+;;;###autoload
 (defun haskell-ds-create-imenu-index ()
   "Function for finding `imenu' declarations in Haskell mode.
 Finds all declarations (classes, variables, imports, instances and
@@ -486,7 +515,7 @@ datatypes) in a Haskell file for the `imenu' package."
 	 (index-type-alist '())    ;; Datatypes
 	 ;; Variables for showing progress.
 	 (bufname (buffer-name))
-	 (divisor-of-progress (max 1 (/ (point-max) 100)))
+	 (divisor-of-progress (max 1 (/ (buffer-size) 100)))
 	 ;; The result we wish to return.
 	 result)
     (goto-char (point-min))
@@ -494,7 +523,7 @@ datatypes) in a Haskell file for the `imenu' package."
     ;; starts of the top-level declarations.
     (while (< (point) (point-max))
       (message "Scanning declarations in %s... (%3d%%)" bufname
-	       (/ (point) divisor-of-progress))
+	       (/ (- (point) (point-min)) divisor-of-progress))
       ;; Grab the next declaration.
       (setq result (haskell-ds-generic-find-next-decl bird-literate))
       (if result
@@ -505,13 +534,13 @@ datatypes) in a Haskell file for the `imenu' package."
 		 (start-pos (car posns))
 		 (type (cdr result))
 		 ;; Place `(name . start-pos)' in the correct alist.
-		 (alist (cond
-			 ((eq type 'variable) 'index-var-alist)
-			 ((eq type 'datatype) 'index-type-alist)
-			 ((eq type 'class) 'index-class-alist)
-			 ((eq type 'import) 'index-imp-alist)
-			 ((eq type 'instance) 'index-inst-alist))))
-	    (set alist (cons (cons name start-pos) (eval alist))))))
+		 (sym (cdr (assq type
+                                 '((variable . index-var-alist)
+                                   (datatype . index-type-alist)
+                                   (class . index-class-alist)
+                                   (import . index-imp-alist)
+                                   (instance . index-inst-alist))))))
+	    (set sym (cons (cons name start-pos) (symbol-value sym))))))
     ;; Now sort all the lists, label them, and place them in one list.
     (message "Sorting declarations in %s..." bufname)
     (and index-type-alist
@@ -588,6 +617,7 @@ datatypes) in a Haskell file for the `imenu' package."
   (concat "^" literate-haskell-ds-start-decl-re)
   "As `haskell-ds-func-menu-regexp' but for Bird-style literate scripts.")
 
+(declare-function fume-add-menubar-entry "ext:func-menu")
 (defvar fume-menubar-menu-name)
 (defvar fume-function-name-regexp-alist)
 (defvar fume-find-function-name-method-alist)
@@ -608,7 +638,9 @@ datatypes) in a Haskell file for the `imenu' package."
   (local-set-key [(meta button1)] 'fume-mouse-function-goto))
 
 ;; The main functions to turn on declaration scanning.
+;;;###autoload
 (defun turn-on-haskell-decl-scan ()
+  (interactive)
   "Unconditionally activate `haskell-decl-scan-mode'."
   (haskell-decl-scan-mode 1))
 
@@ -657,6 +689,7 @@ is nil or `tex', a non-literate or LaTeX-style literate script is
 assumed, respectively.
 
 Invokes `haskell-decl-scan-mode-hook'."
+  (interactive)
   (if (boundp 'beginning-of-defun-function)
       (if haskell-decl-scan-mode
           (progn
@@ -682,5 +715,8 @@ Invokes `haskell-decl-scan-mode-hook'."
 
 (provide 'haskell-decl-scan)
 
-;; arch-tag: f4335fd8-4b6c-472e-9899-004d47d94818
+;; Local Variables:
+;; byte-compile-warnings: (not cl-functions)
+;; End:
+
 ;;; haskell-decl-scan.el ends here
