@@ -17,9 +17,7 @@
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs; see the file COPYING.  If not, write to the
-;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-;; Boston, MA 02111-1307, USA.
+;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 ;;; Commentary:
 
@@ -35,10 +33,24 @@
 
 ;;; Code:
 
-(require 'syntax nil t)			; Emacs 21 add-on
+(require 'syntax)
+(with-no-warnings (require 'cl))
+
+;; Dynamically scoped variables.
+(defvar following-token)
+(defvar current-token)
+(defvar left-indent)
+(defvar starter-indent)
+(defvar current-indent)
+(defvar layout-indent)
+(defvar parse-line-number)
+(defvar possible-indentations)
+(defvar indentation-point)
+(defvar haskell-literate)
 
 (defgroup haskell-indentation nil
   "Haskell indentation."
+  :link '(custom-manual "(haskell-mode)Indentation")
   :group 'haskell
   :prefix "haskell-indentation-")
 
@@ -47,6 +59,16 @@
   :type 'boolean
   :group 'haskell-indentation)
 
+(defcustom haskell-indentation-delete-backward-indentation t
+  "Delete backward removes indentation."
+  :type 'boolean
+  :group 'haskell-indentation)
+
+(defcustom haskell-indentation-delete-indentation t
+  "Delete removes indentation."
+  :type 'boolean
+  :group 'haskell-indentation)
+
 (defcustom haskell-indentation-layout-offset 2
   "Extra indentation to add before expressions in a haskell layout list."
   :type 'integer
@@ -67,6 +89,22 @@
   :type 'integer
   :group 'haskell-indentation)
 
+(defcustom haskell-indentation-where-pre-offset 2
+  "Extra indentation before the keyword `where'."
+  :type 'integer
+  :group 'haskell-indentation)
+
+(defcustom haskell-indentation-where-post-offset 2
+  "Extra indentation after the keyword `where'."
+  :type 'integer
+  :group 'haskell-indentation)
+
+(defcustom haskell-indentation-birdtrack-extra-space t
+  "Append a space after every birdtrack in literate mode."
+  :type 'boolean
+  :group 'haskell-indentation)
+
+
 ;; Avoid a global bogus definition (which the original run-time
 ;; `defun' made), and support Emacs 21 without the syntax.el add-on.
 (eval-when-compile
@@ -81,6 +119,9 @@
     (define-key keymap [?\C-d] 'haskell-indentation-delete-char)
     keymap))
 
+;; Used internally
+(defvar haskell-indent-last-position)
+
 ;;;###autoload
 (define-minor-mode haskell-indentation-mode
   "Haskell indentation mode that deals with the layout rule.
@@ -100,6 +141,7 @@ autofill-mode."
     (set (make-local-variable 'haskell-indent-last-position)
          nil)))
 
+;;;###autoload
 (defun turn-on-haskell-indentation ()
   "Turn on the haskell-indentation minor mode."
   (interactive)
@@ -120,11 +162,25 @@ autofill-mode."
       ,except
       (message "%s" (cdr parse-error-string)))))
 
+(defun haskell-current-column ()
+  "Compute current column according to haskell syntax rules,
+  correctly ignoring composition."
+  (save-excursion
+    (let ((start (point))
+          (cc 0))
+      (beginning-of-line)
+      (while (< (point) start)
+        (if (= (char-after) ?\t)
+            (setq cc (* 8 (+ 1 (/ cc 8))))
+          (incf cc))
+        (forward-char))
+      cc)))
+
 (defun kill-indented-line (&optional arg)
   "`kill-line' for indented text.
 Preserves indentation and removes extra whitespace"
   (interactive "P")
-  (let ((col (current-column))
+  (let ((col (haskell-current-column))
 	(old-point (point)))
     (cond ((or (and (numberp arg) (< arg 0))
 	       (and (not (looking-at "[ \t]*$"))
@@ -144,7 +200,7 @@ Preserves indentation and removes extra whitespace"
 			  (forward-line arg)
 			  (point)))
 	   (skip-chars-forward " \t")
-	   (if (> (current-column) col)
+	   (if (> (haskell-current-column) col)
 	       (move-to-column col)))
 	  (t				; killing from not empty line:
 					; kill all indentation
@@ -156,39 +212,74 @@ Preserves indentation and removes extra whitespace"
 			  (point)))))))
 
 (defun haskell-indentation-auto-fill-function ()
-  (when (> (current-column) fill-column)
-    (while (> (current-column) fill-column)
+  (when (> (haskell-current-column) fill-column)
+    (while (> (haskell-current-column) fill-column)
       (skip-syntax-backward "-")
       (skip-syntax-backward "^-"))
     (let ((auto-fill-function nil)
 	  (indent (car (last (haskell-indentation-find-indentations)))))
       (newline)
+      (when (eq haskell-literate 'bird)
+	(insert ">"))
       (indent-to indent)
       (end-of-line))))
 
 (defun haskell-indentation-reindent (col)
   (beginning-of-line)
   (delete-region (point)
-		 (progn (skip-syntax-forward "-")
-			(point)))
+		 (progn
+		   (when (and (eq haskell-literate 'bird)
+			      (eq (char-after) ?>))
+		     (forward-char))
+		   (skip-syntax-forward "-")
+		   (point)))
+  (when (eq haskell-literate 'bird)
+    (insert ">"))
   (indent-to col))
 
+(defun haskell-indentation-current-indentation ()
+  (if (eq haskell-literate 'bird)
+      (save-excursion
+	(beginning-of-line)
+	(forward-char)
+	(skip-syntax-forward "-")
+	(current-column))
+    (current-indentation)))
+
+(defun haskell-indentation-outside-bird-line () 
+  (and (eq haskell-literate 'bird)
+       (or (< (current-column) 2)
+	   (save-excursion
+	     (beginning-of-line)
+	     (not (eq (char-after) ?>))))))
+
 (defun haskell-newline-and-indent ()
   (interactive)
-  (on-parse-error (newline)
-     (let* ((cc (current-column))
-            (ci (current-indentation))
+  (if (haskell-indentation-outside-bird-line)
+      (progn
+	(delete-horizontal-space)
+	(newline))
+    (on-parse-error
+     (newline)
+     (let* ((cc (haskell-current-column))
+            (ci (haskell-indentation-current-indentation))
             (indentations (haskell-indentation-find-indentations)))
        (skip-syntax-forward "-")
        (if (prog1 (and (eolp)
-                       (not (= (current-column) ci)))
-             (newline))
+                       (not (= (haskell-current-column) ci)))
+	     (delete-horizontal-space)
+	     (if (not (eq haskell-literate 'bird))
+		 (newline)
+	       (when haskell-indentation-birdtrack-extra-space
+		 (indent-to 2))
+	       (newline)
+	       (insert "> ")))
            (haskell-indentation-reindent
             (max (haskell-indentation-butlast indentations)
                  (haskell-indentation-matching-indentation
                   ci indentations)))
          (haskell-indentation-reindent (haskell-indentation-matching-indentation
-                                        cc indentations))))))
+                                        cc indentations)))))))
 
 (defun haskell-indentation-one-indentation (col indentations)
   (let* ((last-pair (last indentations)))
@@ -242,16 +333,16 @@ Preserves indentation and removes extra whitespace"
   (when (save-excursion
 	  (beginning-of-line)
 	  (not (nth 8 (syntax-ppss))))
-    (let ((ci (current-indentation))
-          (start-column (current-column)))
-      (cond ((> (current-column) ci)
+    (let ((ci (haskell-indentation-current-indentation))
+          (start-column (haskell-current-column)))
+      (cond ((> (haskell-current-column) ci)
 	     (save-excursion
 	       (move-to-column ci)
 	       (haskell-indentation-reindent
 		(haskell-indentation-one-indentation
 		 ci (haskell-indentation-find-indentations)))))
 
-	    ((= (current-column) ci)
+	    ((= (haskell-current-column) ci)
 	     (haskell-indentation-reindent
 	      (haskell-indentation-next-indentation
 	       ci (haskell-indentation-find-indentations))))
@@ -260,7 +351,7 @@ Preserves indentation and removes extra whitespace"
 	       (haskell-indentation-reindent
 		(haskell-indentation-matching-indentation
 		 ci (haskell-indentation-find-indentations)))))
-      (cond ((not (= (current-column) start-column))
+      (cond ((not (= (haskell-current-column) start-column))
              (setq haskell-indent-last-position nil))
             ((not haskell-indentation-cycle-warn)
              (haskell-indentation-reindent
@@ -278,79 +369,99 @@ Preserves indentation and removes extra whitespace"
 
 (defun haskell-indentation-delete-backward-char (n)
   (interactive "p")
-  (on-parse-error (backward-delete-char 1)
-     (cond
-      ((and delete-selection-mode
-            mark-active
-            (not (= (point) (mark))))
-       (delete-region (mark) (point)))
-      ((or (= (current-column) 0)
-           (> (current-column) (current-indentation))
-           (nth 8 (syntax-ppss)))
-       (delete-backward-char n))
-      (t (let* ((ci (current-indentation))
-                (pi (haskell-indentation-previous-indentation
-                     ci (haskell-indentation-find-indentations))))
-           (save-excursion
-             (cond (pi
-                    (move-to-column pi)
-                    (delete-region (point)
-                                   (progn (move-to-column ci)
-                                          (point))))
-                   (t
-                    (beginning-of-line)
-                    (delete-region (max (point-min) (- (point) 1))
-                                   (progn (move-to-column ci)
-                                          (point)))))))))))
+  (on-parse-error
+   (delete-char (- n))
+   (cond
+    ((haskell-indentation-outside-bird-line)
+     (delete-char (- n)))
+    ((and delete-selection-mode
+	  mark-active
+	  (not (= (point) (mark))))
+     (delete-region (mark) (point)))
+    ((or (= (haskell-current-column) 0)
+	 (> (haskell-current-column) (haskell-indentation-current-indentation))
+	(nth 8 (syntax-ppss)))
+     (delete-char (- n)))
+    (haskell-indentation-delete-backward-indentation
+       (let* ((ci (haskell-indentation-current-indentation))
+	      (pi (haskell-indentation-previous-indentation
+		   ci (haskell-indentation-find-indentations))))
+	 (save-excursion
+	   (cond (pi
+		  (move-to-column pi)
+		  (delete-region (point)
+				 (progn (move-to-column ci)
+					(point))))
+		 (t
+		  (beginning-of-line)
+		  (delete-region (max (point-min) (- (point) 1))
+				 (progn (move-to-column ci)
+					(point))))))))
+    (t (delete-char (- n))))))
 
 (defun haskell-indentation-delete-char (n)
   (interactive "p")
-  (on-parse-error (delete-char 1)
-    (cond
-     ((and delete-selection-mode
-           mark-active
-           (not (= (point) (mark))))
-      (delete-region (mark) (point)))
-     ((or (eolp)
-          (>= (current-column) (current-indentation))
-          (nth 8 (syntax-ppss)))
-      (delete-char n))
-     (t
-      (let* ((ci (current-indentation))
-             (pi (haskell-indentation-previous-indentation
-                  ci (haskell-indentation-find-indentations))))
-        (save-excursion
-          (if (and pi (> pi (current-column)))
-              (move-to-column pi))
-          (delete-region (point)
-                         (progn (move-to-column ci)
-                                (point)))))))))
+  (if (haskell-indentation-outside-bird-line)
+      (delete-char n)
+    (on-parse-error (delete-char n)
+     (cond
+      ((and delete-selection-mode
+	    mark-active
+	    (not (= (point) (mark))))
+       (delete-region (mark) (point)))
+      ((and (eq haskell-literate 'bird)
+	    (looking-at "\n> "))
+       (delete-char (+ n 2)))
+      ((or (eolp)
+	   (>= (haskell-current-column) (haskell-indentation-current-indentation))
+	   (nth 8 (syntax-ppss)))
+       (delete-char n))
+      (haskell-indentation-delete-indentation
+       (let* ((ci (haskell-indentation-current-indentation))
+	      (pi (haskell-indentation-previous-indentation
+		   ci (haskell-indentation-find-indentations))))
+	 (save-excursion
+	   (if (and pi (> pi (haskell-current-column)))
+	       (move-to-column pi))
+	   (delete-region (point)
+			  (progn (move-to-column ci)
+				 (point))))))
+    (t (delete-char (- n)))))))
 
 (defun haskell-indentation-goto-least-indentation ()
   (beginning-of-line)
-  (catch 'return
-    (while (not (bobp))
-      (forward-comment (- (buffer-size)))
-      (beginning-of-line)
-      (let ((ps (nth 8 (syntax-ppss))))
-		(when ps ;; inside comment or string
-		  (goto-char ps)))
-      (when (= 0 (current-indentation))
-		(throw 'return nil))))
-  (beginning-of-line)
-  (when (bobp)
-    (forward-comment (buffer-size))))
-
-;; Dynamically scoped variables.
-(defvar following-token)
-(defvar current-token)
-(defvar left-indent)
-(defvar starter-indent)
-(defvar current-indent)
-(defvar layout-indent)
-(defvar parse-line-number)
-(defvar possible-indentations)
-(defvar indentation-point)
+  (if (eq haskell-literate 'bird)
+      (catch 'return
+	(while t
+	  (when (not (eq (char-after) ?>))
+	    (forward-line)
+	    (forward-char 2)
+	    (throw 'return nil))
+	  (let ((ps (nth 8 (syntax-ppss))))
+	    (when ps ;; inside comment or string
+	      (goto-char ps)
+	      (beginning-of-line)))
+	  (when (and (>= 2 (haskell-indentation-current-indentation))
+		     (not (looking-at ">\\s-*$")))
+	    (forward-char 2)
+	    (throw 'return nil))
+	  (when (bobp)
+	    (forward-char 2)
+	    (throw 'return nil))
+	  (forward-line -1)))
+    ;; not bird style
+    (catch 'return
+      (while (not (bobp))
+	(forward-comment (- (buffer-size)))
+	(beginning-of-line)
+	(let ((ps (nth 8 (syntax-ppss))))
+	  (when ps ;; inside comment or string
+	    (goto-char ps)))
+	(when (= 0 (haskell-indentation-current-indentation))
+	  (throw 'return nil))))
+    (beginning-of-line)
+    (when (bobp)
+      (forward-comment (buffer-size)))))
 
 (defun haskell-indentation-parse-to-indentations ()
   (save-excursion
@@ -367,18 +478,22 @@ Preserves indentation and removes extra whitespace"
 	  possible-indentations)
       (haskell-indentation-goto-least-indentation)
       (if (<= indentation-point (point))
-	  '(0)
+	  (haskell-indentation-first-indentation)
 	(setq current-token (haskell-indentation-peek-token))
 	(catch 'parse-end
 	  (haskell-indentation-toplevel)
-	  (when (not (equal current-token 'end-tokens))
+	  (unless (eq current-token 'end-tokens)
 	    (parse-error "Illegal token: %s" current-token)))
 	possible-indentations))))
 
+(defun haskell-indentation-first-indentation ()
+  (if (eq haskell-literate 'bird) '(2) '(0)))
+
 (defun haskell-indentation-find-indentations ()
   (let ((ppss (syntax-ppss)))
     (cond
-     ((nth 3 ppss) '(0))
+     ((nth 3 ppss)
+      (haskell-indentation-first-indentation))
      ((nth 4 ppss)
       (if (save-excursion
 	    (and (skip-syntax-forward "-")
@@ -386,20 +501,34 @@ Preserves indentation and removes extra whitespace"
 		 (not (> (forward-line 1) 0))
 		 (not (nth 4 (syntax-ppss)))))
 	  (haskell-indentation-parse-to-indentations)
-	'(0)))
+	(haskell-indentation-first-indentation)))
      (t
       (haskell-indentation-parse-to-indentations)))))
 
+(defconst haskell-indentation-unicode-tokens
+  '(("→" . "->")     ;; #x2192 RIGHTWARDS ARROW
+    ("∷" . "::")     ;; #x2237 PROPORTION
+    ("←" . "<-")     ;; #x2190 LEFTWARDS ARROW
+    ("⇒" . "=>")     ;; #x21D2 RIGHTWARDS DOUBLE ARROW
+    ("∀" . "forall") ;; #x2200 FOR ALL
+    ("↢" . "-<")     ;; #x2919 LEFTWARDS ARROW-TAIL
+    ("↣" . ">-")     ;; #x291A RIGHTWARDS ARROW-TAIL
+    ("⤛" . "-<<")    ;; #x291B LEFTWARDS DOUBLE ARROW-TAIL
+    ("⤜" . ">>-")    ;; #x291C RIGHTWARDS DOUBLE ARROW-TAIL
+    ("★" . "*"))     ;; #x2605 BLACK STAR
+  "Translation dictionary from UnicodeSyntax tokens to their ASCII representation.")
+
 (defconst haskell-indentation-toplevel-list
   '(("module" . haskell-indentation-module)
-    ("data" . haskell-indentation-data)
-    ("type" . haskell-indentation-data)
-    ("newtype" . haskell-indentation-data)
+    ("data" . (lambda () (haskell-indentation-statement-right #'haskell-indentation-data)))
+    ("type" . (lambda () (haskell-indentation-statement-right #'haskell-indentation-data)))
+    ("newtype" . (lambda () (haskell-indentation-statement-right #'haskell-indentation-data)))
     ("class" . haskell-indentation-class-declaration)
     ("instance" . haskell-indentation-class-declaration )))
 
 (defconst haskell-indentation-type-list
-  '(("::"    . (lambda () (haskell-indentation-statement-right #'haskell-indentation-type)))
+  '(("::"    . (lambda () (haskell-indentation-with-starter
+			   (lambda () (haskell-indentation-separated #'haskell-indentation-type "->" nil)) nil)))
     ("("     . (lambda () (haskell-indentation-list #'haskell-indentation-type
 						    ")" "," nil)))
     ("["     . (lambda () (haskell-indentation-list #'haskell-indentation-type
@@ -422,15 +551,21 @@ Preserves indentation and removes extra whitespace"
 			   #'haskell-indentation-expression-layout nil)))
     ("mdo"   . (lambda () (haskell-indentation-with-starter
 			   #'haskell-indentation-expression-layout nil)))
+    ("rec"   . (lambda () (haskell-indentation-with-starter
+			   #'haskell-indentation-expression-layout nil)))
     ("case"  . (lambda () (haskell-indentation-phrase
 			   '(haskell-indentation-expression
 			     "of" haskell-indentation-case-layout))))
     ("\\"    . (lambda () (haskell-indentation-phrase
 			   '(haskell-indentation-expression
 			     "->" haskell-indentation-expression))))
+    ("proc"  . (lambda () (haskell-indentation-phrase
+			   '(haskell-indentation-expression
+			     "->" haskell-indentation-expression))))
     ("where" . (lambda () (haskell-indentation-with-starter
-			   #'haskell-indentation-declaration-layout nil)))
-    ("::"    . (lambda () (haskell-indentation-statement-right #'haskell-indentation-type)))
+			   #'haskell-indentation-declaration-layout nil t)))
+    ("::"    . (lambda () (haskell-indentation-with-starter
+			   (lambda () (haskell-indentation-separated #'haskell-indentation-type "->" nil)) nil)))
     ("="     . (lambda () (haskell-indentation-statement-right #'haskell-indentation-expression)))
     ("<-"    . (lambda () (haskell-indentation-statement-right #'haskell-indentation-expression)))
     ("("     . (lambda () (haskell-indentation-list #'haskell-indentation-expression
@@ -456,52 +591,52 @@ Preserves indentation and removes extra whitespace"
    nil))
 
 (defun haskell-indentation-fundep1 ()
-  (let ((current-indent (current-column)))
+  (let ((current-indent (haskell-current-column)))
     (while (member current-token '(value "->"))
       (haskell-indentation-read-next-token))
-    (when (and (equal current-token 'end-tokens)
+    (when (and (eq current-token 'end-tokens)
 	       (member following-token '(value "->")))
       (haskell-indentation-add-indentation current-indent))))
 
 (defun haskell-indentation-toplevel ()
   (haskell-indentation-layout
    (lambda ()
-	 (let ((parser (assoc current-token haskell-indentation-toplevel-list)))
-	   (if parser
-		   (funcall (cdr parser))
-		 (haskell-indentation-declaration))))))
+     (let ((parser (assoc current-token haskell-indentation-toplevel-list)))
+       (if parser
+	   (funcall (cdr parser))
+	 (haskell-indentation-declaration))))))
 
 (defun haskell-indentation-type ()
-  (let ((current-indent (current-column)))
+  (let ((current-indent (haskell-current-column)))
     (catch 'return
       (while t
-		(cond
-		 ((member current-token '(value operator "->"))
-		  (haskell-indentation-read-next-token))
-
-		 ((equal current-token 'end-tokens)
-		  (when (member following-token
-						'(value operator no-following-token
-								"->" "(" "[" "{" "::"))
-			(haskell-indentation-add-indentation current-indent))
-		  (throw 'return nil))
-		 
-		 (t (let ((parser (assoc current-token haskell-indentation-type-list)))
-			  (if (not parser)
-				  (throw 'return nil)
-				(funcall (cdr parser))))))))))
+	(cond
+	 ((member current-token '(value operator "->"))
+	  (haskell-indentation-read-next-token))
+	 
+	 ((eq current-token 'end-tokens)
+	  (when (member following-token
+			'(value operator no-following-token
+				"->" "(" "[" "{" "::"))
+	    (haskell-indentation-add-indentation current-indent))
+	  (throw 'return nil))
+	 
+	 (t (let ((parser (assoc current-token haskell-indentation-type-list)))
+	      (if (not parser)
+		  (throw 'return nil)
+		(funcall (cdr parser))))))))))
 
 (defun haskell-indentation-data ()
   (haskell-indentation-with-starter
    (lambda ()
-     (when (equal current-token "instance")
+     (when (string= current-token "instance")
        (haskell-indentation-read-next-token))
      (haskell-indentation-type)
-     (cond ((equal current-token "=")
+     (cond ((string= current-token "=")
 	    (haskell-indentation-with-starter
 	     (lambda () (haskell-indentation-separated #'haskell-indentation-type "|" "deriving"))
 	     nil))
-	   ((equal current-token "where")
+	   ((string= current-token "where")
 	    (haskell-indentation-with-starter
 	     #'haskell-indentation-expression-layout nil))))
    nil))
@@ -510,9 +645,9 @@ Preserves indentation and removes extra whitespace"
   (haskell-indentation-with-starter
    (lambda ()
      (haskell-indentation-type)
-     (when (equal current-token "|")
+     (when (string= current-token "|")
        (haskell-indentation-fundep))
-     (when (equal current-token "where")
+     (when (string= current-token "where")
        (haskell-indentation-with-starter
 	#'haskell-indentation-expression-layout nil)))
    nil))
@@ -520,111 +655,113 @@ Preserves indentation and removes extra whitespace"
 (defun haskell-indentation-module ()
   (haskell-indentation-with-starter
    (lambda ()
-	 (let ((current-indent (current-column)))
-	   (haskell-indentation-read-next-token)
-	   (when (equal current-token "(")
-		 (haskell-indentation-list
-		  #'haskell-indentation-module-export
-		  ")" "," nil))
-	   (when (equal current-token 'end-tokens)
-		 (haskell-indentation-add-indentation current-indent)
-		 (throw 'parse-end nil))
-	   (when (equal current-token "where")
-		 (haskell-indentation-read-next-token)
-		 (when (equal current-token 'end-tokens)
-		   (haskell-indentation-add-layout-indent)
-		   (throw 'parse-end nil))
-		 (haskell-indentation-layout #'haskell-indentation-toplevel))))
+     (let ((current-indent (haskell-current-column)))
+       (haskell-indentation-read-next-token)
+       (when (string= current-token "(")
+	 (haskell-indentation-list
+	  #'haskell-indentation-module-export
+	  ")" "," nil))
+       (when (eq current-token 'end-tokens)
+	 (haskell-indentation-add-indentation current-indent)
+	 (throw 'parse-end nil))
+       (when (string= current-token "where")
+	 (haskell-indentation-read-next-token)
+	 (when (eq current-token 'end-tokens)
+	   (haskell-indentation-add-layout-indent)
+	   (throw 'parse-end nil))
+	 (haskell-indentation-layout #'haskell-indentation-toplevel))))
    nil))
 
 (defun haskell-indentation-module-export ()
-  (cond ((equal current-token "module")
-		 (let ((current-indent (current-column)))
-		   (haskell-indentation-read-next-token)
-		   (cond ((equal current-token 'end-tokens)
-				  (haskell-indentation-add-indentation current-indent))
-				 ((equal current-token 'value)
-				  (haskell-indentation-read-next-token)))))
-		(t (haskell-indentation-type))))
+  (cond ((string= current-token "module")
+	 (let ((current-indent (haskell-current-column)))
+	   (haskell-indentation-read-next-token)
+	   (cond ((eq current-token 'end-tokens)
+		  (haskell-indentation-add-indentation current-indent))
+		 ((eq current-token 'value)
+		  (haskell-indentation-read-next-token)))))
+	(t (haskell-indentation-type))))
 
 (defun haskell-indentation-list (parser end sep stmt-sep)
   (haskell-indentation-with-starter
    `(lambda () (haskell-indentation-separated #',parser
-											  ,sep
-											  ,stmt-sep))
+					      ,sep
+					      ,stmt-sep))
    end))
 
-(defun haskell-indentation-with-starter (parser end)
-  (let ((starter-column (current-column))
-		(current-indent current-indent)
-		(left-indent (if (= (current-column) (current-indentation))
-						 (current-column) left-indent)))
+(defun haskell-indentation-with-starter (parser end &optional where-expr?)
+  (let ((starter-column (haskell-current-column))
+	(current-indent current-indent)
+	(left-indent (if (= (haskell-current-column) (haskell-indentation-current-indentation))
+			 (haskell-current-column) left-indent)))
     (haskell-indentation-read-next-token)
-    (when (equal current-token 'end-tokens)
+    (when (eq current-token 'end-tokens)
       (if (equal following-token end)
 	  (haskell-indentation-add-indentation starter-column)
+        (if where-expr?
+            (haskell-indentation-add-where-post-indent left-indent)
 	  (haskell-indentation-add-indentation
-	   (+ left-indent haskell-indentation-left-offset)))
+	   (+ left-indent haskell-indentation-left-offset))))
       (throw 'parse-end nil))
-    (let* ((current-indent (current-column))
-		   (starter-indent (min starter-column current-indent))
-		   (left-indent (if end (+ current-indent haskell-indentation-starter-offset)
-						  left-indent)))
+    (let* ((current-indent (haskell-current-column))
+	   (starter-indent (min starter-column current-indent))
+	   (left-indent (if end (+ current-indent haskell-indentation-starter-offset)
+			  left-indent)))
       (funcall parser)
-      (cond ((equal current-token 'end-tokens)
-			 (when (equal following-token end)
-			   (haskell-indentation-add-indentation starter-indent))
-			 (when end (throw 'parse-end nil))) ;; add no indentations
-			((equal current-token end)
-			 (haskell-indentation-read-next-token)) ;; continue
-			(end (parse-error "Illegal token: %s" current-token))))))
+      (cond ((eq current-token 'end-tokens)
+	     (when (equal following-token end)
+	       (haskell-indentation-add-indentation starter-indent))
+	     (when end (throw 'parse-end nil))) ;; add no indentations
+	    ((equal current-token end)
+	     (haskell-indentation-read-next-token)) ;; continue
+	    (end (parse-error "Illegal token: %s" current-token))))))
 
 (defun haskell-indentation-case ()
   (haskell-indentation-expression)
-  (cond ((equal current-token 'end-tokens)
+  (cond ((eq current-token 'end-tokens)
 	 (haskell-indentation-add-indentation current-indent))
-	((equal current-token "|")
+	((string= current-token "|")
 	 (haskell-indentation-with-starter
 	  (lambda () (haskell-indentation-separated #'haskell-indentation-case "|" nil))
 	  nil))
-	((equal current-token "->")
+	((string= current-token "->")
 	 (haskell-indentation-statement-right #'haskell-indentation-expression))
 	;; otherwise fallthrough
 	))
 
 (defun haskell-indentation-statement-right (parser)
     (haskell-indentation-read-next-token)
-    (when (equal current-token 'end-tokens)
+    (when (eq current-token 'end-tokens)
       (haskell-indentation-add-indentation
        (+ left-indent haskell-indentation-left-offset))
       (throw 'parse-end nil))
-    (let ((current-indent (current-column)))
+    (let ((current-indent (haskell-current-column)))
 	  (funcall parser)))
 
 (defun haskell-indentation-simple-declaration ()
   (haskell-indentation-expression)
-  (cond ((equal current-token "=")
+  (cond ((string= current-token "=")
 	 (haskell-indentation-statement-right #'haskell-indentation-expression))
-	((equal current-token "::")
+	((string= current-token "::")
 	 (haskell-indentation-statement-right #'haskell-indentation-type))
-	((and (equal current-token 'end-tokens)
-	      (equal following-token "="))
+	((and (eq current-token 'end-tokens)
+	      (string= following-token "="))
 	 (haskell-indentation-add-indentation current-indent)
 	 (throw 'parse-end nil))))
 
 (defun haskell-indentation-declaration ()
   (haskell-indentation-expression)
-  (cond ((equal current-token "|")
+  (cond ((string= current-token "|")
 	 (haskell-indentation-with-starter
 	  (lambda () (haskell-indentation-separated #'haskell-indentation-expression "," "|"))
 	  nil))
-	((equal current-token 'end-tokens)
+	((eq current-token 'end-tokens)
 	 (when (member following-token '("|" "=" "::" ","))
 	   (haskell-indentation-add-indentation current-indent)
 	   (throw 'parse-end nil)))))
 
 (defun haskell-indentation-layout (parser)
-  (if (equal current-token "{")
+  (if (string= current-token "{")
       (haskell-indentation-list parser "}" ";" nil)
     (haskell-indentation-implicit-layout-list parser)))
 
@@ -633,17 +770,16 @@ Preserves indentation and removes extra whitespace"
 		  value operator no-following-token)))
 
 (defun haskell-indentation-expression ()
-  (let ((current-indent (current-column)))
+  (let ((current-indent (haskell-current-column)))
     (catch 'return
       (while t
 	(cond
-	 ((member current-token '(value operator))
+	 ((memq current-token '(value operator))
 	  (haskell-indentation-read-next-token))
 
-	 ((equal current-token 'end-tokens)
-	  (cond ((equal following-token "where")
-		 (haskell-indentation-add-indentation
-		  (+ left-indent haskell-indentation-left-offset)))
+	 ((eq current-token 'end-tokens)
+	  (cond ((string= following-token "where")
+		 (haskell-indentation-add-where-pre-indent))
 		((haskell-indentation-expression-token following-token)
 		 (haskell-indentation-add-indentation
 		  current-indent)))
@@ -653,8 +789,8 @@ Preserves indentation and removes extra whitespace"
 	      (when (null parser)
 		(throw 'return nil))
 	      (funcall (cdr parser))
-	      (when (and (equal current-token 'end-tokens)
-			 (equal (car parser) "let")
+	      (when (and (eq current-token 'end-tokens)
+			 (string= (car parser) "let")
 			 (= haskell-indentation-layout-offset current-indent)
 			 (haskell-indentation-expression-token following-token))
 		;; inside a layout, after a let construct
@@ -686,10 +822,10 @@ Preserves indentation and removes extra whitespace"
 	     (haskell-indentation-at-separator))
 
 	    ((equal current-token stmt-separator)
-	     (setq starter-indent (current-column))
+	     (setq starter-indent (haskell-current-column))
 	     (haskell-indentation-at-separator))
 
-	    ((equal current-token 'end-tokens)
+	    ((eq current-token 'end-tokens)
 	     (cond ((or (equal following-token separator)
 			(equal following-token stmt-separator))
 		    (haskell-indentation-add-indentation starter-indent)
@@ -700,29 +836,29 @@ Preserves indentation and removes extra whitespace"
 
 (defun haskell-indentation-at-separator ()
   (let ((separator-column
-	 (and (= (current-column) (current-indentation))
-	      (current-column))))
+	 (and (= (haskell-current-column) (haskell-indentation-current-indentation))
+	      (haskell-current-column))))
     (haskell-indentation-read-next-token)
     (cond ((eq current-token 'end-tokens)
 	   (haskell-indentation-add-indentation current-indent)
 	   (throw 'return nil))
 	  (separator-column ;; on the beginning of the line
-	   (setq current-indent (current-column))
+	   (setq current-indent (haskell-current-column))
 	   (setq starter-indent separator-column)))))
 
 (defun haskell-indentation-implicit-layout-list (parser)
-  (let* ((layout-indent (current-column))
-		 (current-indent (current-column))
-		 (left-indent (current-column)))
+  (let* ((layout-indent (haskell-current-column))
+	 (current-indent (haskell-current-column))
+	 (left-indent (haskell-current-column)))
     (catch 'return
       (while t
 	(let ((left-indent left-indent))
 	  (funcall parser))
 	(cond ((member current-token '(layout-next ";"))
 	       (haskell-indentation-read-next-token))
-	      ((equal current-token 'end-tokens)
+	      ((eq current-token 'end-tokens)
 	       (when (or (haskell-indentation-expression-token following-token)
-					 (equal following-token ";"))
+					 (string= following-token ";"))
 			 (haskell-indentation-add-layout-indent))
 	       (throw 'return nil))
 	      (t (throw 'return nil))))))
@@ -738,15 +874,15 @@ Preserves indentation and removes extra whitespace"
 
 (defun haskell-indentation-phrase-rest (phrase)
   (let ((starter-line parse-line-number))
-    (let ((current-indent (current-column)))
+    (let ((current-indent (haskell-current-column)))
       (funcall (car phrase)))
     (cond
-     ((equal current-token 'end-tokens)
+     ((eq current-token 'end-tokens)
       (cond ((null (cdr phrase))) ;; fallthrough
 	    ((equal following-token (cadr phrase))
 	     (haskell-indentation-add-indentation starter-indent)
 	     (throw 'parse-end nil))
-	    ((equal (cadr phrase) "in")
+	    ((string= (cadr phrase) "in")
 	     (when (= left-indent layout-indent)
 	       (haskell-indentation-add-layout-indent)
 	       (throw 'parse-end nil)))
@@ -755,13 +891,13 @@ Preserves indentation and removes extra whitespace"
      ((null (cdr phrase)))
      
      ((equal (cadr phrase) current-token)
-      (let* ((on-new-line (= (current-column) (current-indentation)))
+      (let* ((on-new-line (= (haskell-current-column) (haskell-indentation-current-indentation)))
 	     (lines-between (- parse-line-number starter-line))
 	     (left-indent (if (<= lines-between 0)
 			      left-indent
 			    starter-indent)))
 	(haskell-indentation-read-next-token)
-	(when (equal current-token 'end-tokens)
+	(when (eq current-token 'end-tokens)
 	  (haskell-indentation-add-indentation
 	   (cond ((member (cadr phrase) '("then" "else"))
 		  (+ starter-indent haskell-indentation-ifte-offset))
@@ -774,7 +910,7 @@ Preserves indentation and removes extra whitespace"
 	  (throw 'parse-end nil))
 	(haskell-indentation-phrase-rest (cddr phrase))))
 
-     ((equal (cadr phrase) "in")) ;; fallthrough
+     ((string= (cadr phrase) "in")) ;; fallthrough
      (t (parse-error "Expecting %s" (cadr phrase))))))
 
 (defun haskell-indentation-add-indentation (indent)
@@ -786,6 +922,17 @@ Preserves indentation and removes extra whitespace"
 (defun haskell-indentation-add-layout-indent ()
   (haskell-indentation-push-indentation layout-indent))
 
+(defun haskell-indentation-add-where-pre-indent ()
+  (haskell-indentation-push-indentation
+   (+ layout-indent haskell-indentation-where-pre-offset))
+  (if (= layout-indent haskell-indentation-layout-offset)
+      (haskell-indentation-push-indentation
+       haskell-indentation-where-pre-offset)))
+
+(defun haskell-indentation-add-where-post-indent (indent)
+  (haskell-indentation-push-indentation
+   (+ indent haskell-indentation-where-post-offset)))
+
 (defun haskell-indentation-push-indentation (indent)
   (when (or (null possible-indentations)
 	    (< indent (car possible-indentations)))
@@ -796,6 +943,7 @@ Preserves indentation and removes extra whitespace"
   (let ((current-token nil)
 	(following-token nil)
 	(layout-indent 0)
+	(parse-line-number 0)
 	(indentation-point (mark)))
     (haskell-indentation-read-next-token)))
 
@@ -803,15 +951,15 @@ Preserves indentation and removes extra whitespace"
   (cond ((eq current-token 'end-tokens)
 	 'end-tokens)
 	((eq current-token 'layout-end)
-	 (cond ((> layout-indent (current-column))
+	 (cond ((> layout-indent (haskell-current-column))
 		'layout-end)
-	       ((= layout-indent (current-column))
+	       ((= layout-indent (haskell-current-column))
 		(setq current-token 'layout-next))
-	       ((< layout-indent (current-column))
+	       ((< layout-indent (haskell-current-column))
 		(setq current-token (haskell-indentation-peek-token)))))
 	((eq current-token 'layout-next)
 	 (setq current-token (haskell-indentation-peek-token)))
-	((> layout-indent (current-column))
+	((> layout-indent (haskell-current-column))
 	 (setq current-token 'layout-end))
 	(t
 	 (haskell-indentation-skip-token)
@@ -822,24 +970,28 @@ Preserves indentation and removes extra whitespace"
 			 (haskell-indentation-peek-token)
 		       'no-following-token))
 	       (setq current-token 'end-tokens))
-	   (when (= (current-column) (current-indentation))
+	   (when (= (haskell-current-column) (haskell-indentation-current-indentation))
 	     ;; on a new line
-	     (setq current-indent (current-column))
-	     (setq left-indent (current-column))
+	     (setq current-indent (haskell-current-column))
+	     (setq left-indent (haskell-current-column))
 	     (setq parse-line-number (+ parse-line-number 1)))
-	   (cond ((> layout-indent (current-column))
+	   (cond ((> layout-indent (haskell-current-column))
 		  (setq current-token 'layout-end))
-		 ((= layout-indent (current-column))
+		 ((= layout-indent (haskell-current-column))
 		  (setq current-token 'layout-next))
 		 (t (setq current-token (haskell-indentation-peek-token))))))))
 
 (defun haskell-indentation-peek-token ()
-  (cond ((looking-at "\\(if\\|then\\|else\\|let\\|in\\|mdo\\|do\\|case\\|of\\|where\\|module\\|deriving\\|data\\|type\\|newtype\\|class\\|instance\\)\\([^[:alpha:]']\\|$\\)")
-	 (match-string 1))
+  "Return token starting at point."
+  (cond ((looking-at "\\(if\\|then\\|else\\|let\\|in\\|mdo\\|rec\\|do\\|proc\\|case\\|of\\|where\\|module\\|deriving\\|data\\|type\\|newtype\\|class\\|instance\\)\\([^[:alnum:]'_]\\|$\\)")
+	 (match-string-no-properties 1))
 	((looking-at "[][(){}[,;]")
-	 (match-string 0))
+	 (match-string-no-properties 0))
 	((looking-at "\\(\\\\\\|->\\|<-\\|::\\|=\\||\\)\\([^-:!#$%&*+./<=>?@\\\\^|~]\\|$\\)")
-	 (match-string 1))
+	 (match-string-no-properties 1))
+	((looking-at "\\(→\\|←\\|∷\\)\\([^-:!#$%&*+./<=>?@\\\\^|~]\\|$\\)")
+         (let ((tok (match-string-no-properties 1)))
+           (or (cdr (assoc tok haskell-indentation-unicode-tokens)) tok)))
 	((looking-at"[-:!#$%&*+./<=>?@\\\\^|~`]" )
 	 'operator)
 	(t 'value)))
@@ -847,6 +999,7 @@ Preserves indentation and removes extra whitespace"
 (defun haskell-indentation-skip-token ()
   "Skip to the next token."
   (let ((case-fold-search nil))
+    
     (if (or (looking-at "'\\([^\\']\\|\\\\.\\)*'")
             (looking-at "\"\\([^\\\"]\\|\\\\.\\)*\"")
             (looking-at	; Hierarchical names always start with uppercase
@@ -859,7 +1012,18 @@ Preserves indentation and removes extra whitespace"
         (goto-char (match-end 0))
     ;; otherwise skip until space found
       (skip-syntax-forward "^-"))
-    (forward-comment (buffer-size))))
+    (forward-comment (buffer-size))
+    (while (and (eq haskell-literate 'bird)
+		(bolp)
+		(eq (char-after) ?>))
+      (forward-char)
+      (forward-comment (buffer-size)))))
 
 (provide 'haskell-indentation)
+
+;; Local Variables:
+;; byte-compile-warnings: (not cl-functions)
+;; tab-width: 8
+;; End:
+
 ;;; haskell-indentation.el ends here
