@@ -1,4 +1,4 @@
-;;; haskell-mode.el --- A Haskell editing mode    -*-coding: iso-8859-1;-*-
+;;; haskell-mode.el --- A Haskell editing mode    -*- coding: utf-8 -*-
 
 ;; Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008  Free Software Foundation, Inc
 ;; Copyright (C) 1992, 1997-1998 Simon Marlow, Graeme E Moss, and Tommy Thorn
@@ -9,8 +9,7 @@
 ;;          2001-2002 Reuben Thomas (>=v1.4)
 ;;          2003      Dave Love <fx@gnu.org>
 ;; Keywords: faces files Haskell
-;; Version: v2.8.0
-;; URL: http://www.haskell.org/haskell-mode/
+;; URL: https://github.com/haskell/haskell-mode
 
 ;; This file is not part of GNU Emacs.
 
@@ -25,65 +24,42 @@
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs; see the file COPYING.  If not, write to the
-;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-;; Boston, MA 02111-1307, USA.
+;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-
 ;;; Commentary:
 
-;; Purpose:
+;; A major mode for editing Haskell (the functional programming
+;; language, see URL `http://www.haskell.org') in Emacs.
 ;;
-;; To provide a pleasant mode to browse and edit Haskell files, linking
-;; into the following supported modules:
+;; Some of its major features include:
 ;;
-;; `haskell-font-lock', Graeme E Moss and Tommy Thorn
-;;   Fontifies standard Haskell keywords, symbols, functions, etc.
+;;  - syntax highlighting (font lock),
 ;;
-;; `haskell-decl-scan', Graeme E Moss
-;;   Scans top-level declarations, and places them in a menu.
+;;  - automatic indentation,
 ;;
-;; `haskell-doc', Hans-Wolfgang Loidl
-;;   Echoes types of functions or syntax of keywords when the cursor is idle.
+;;  - on-the-fly documentation,
 ;;
-;; `haskell-indentation', Kristof Bastiaensen
-;;   Intelligent semi-automatic indentation, mark two.
+;;  - interaction with inferior GHCi/Hugs instance,
 ;;
-;; `haskell-indent', Guy Lapalme
-;;   Intelligent semi-automatic indentation.
+;;  - scans declarations and places them in a menu.
 ;;
-;; `haskell-simple-indent', Graeme E Moss and Heribert Schuetz
-;;   Simple indentation.
-;;
-;; `inf-haskell'
-;;   Interaction with an inferior Haskell process.
-;;   It replaces the previous two modules:
-;;     `haskell-hugs', Guy Lapalme
-;;     `haskell-ghci', Chris Web
-;;
-;;
-;; This mode supports full Haskell 1.4 including literate scripts.
-;; In some versions of (X)Emacs it may only support Latin-1, not Unicode.
-;;
-;; History:
+;; See URL `https://github.com/haskell/haskell-mode' and/or
+;; Info node `(haskell-mode)Introduction' for more information.
 ;;
+;; Use `M-x haskell-mode-view-news` (after Haskell Mode is installed)
+;; to show information on recent changes in Haskell Mode.
+
+;;; Change Log:
+
 ;; This mode is based on an editing mode by Simon Marlow 11/1/92
 ;; and heavily modified by Graeme E Moss and Tommy Thorn 7/11/98.
-;; 
-;; If you have any problems or suggestions specific to a supported
-;; module, consult that module for a list of known bugs, and an
-;; author to contact via email.  For general problems or suggestions,
-;; consult the list below, then email gem@cs.york.ac.uk and
-;; thorn@irisa.fr quoting the version of the mode you are using, the
-;; version of Emacs you are using, and a small example of the problem
-;; or suggestion.
 ;;
-;; Version 1.5
+;; Version 1.5:
 ;;   Added autoload for haskell-indentation
 ;;
 ;; Version 1.43:
 ;;   Various tweaks to doc strings and customization support from
-;;   Ville Skytt <scop@xemacs.org>.
+;;   Ville Skyttä <scop@xemacs.org>.
 ;;
 ;; Version 1.42:
 ;;   Added autoload for GHCi inferior mode (thanks to Scott
@@ -143,73 +119,86 @@
 ;;   Altered indent-line-function to indent-relative.
 ;;
 ;; Version 0.100:
-;; 
+;;
 ;;   First official release.
 
-;; Present Limitations/Future Work (contributions are most welcome!):
-;;
-;; . Would like RET in Bird-style literate mode to add a ">" at the
-;;   start of a line when previous line starts with ">".  Or would
-;;   "> " be better?
-;;
-;; . Support for GreenCard?
-;;
-
 ;;; Code:
 
-(eval-when-compile (require 'cl))
-(eval-when-compile
-  ;; Emacs 21 defines `values' as a (run-time) alias for list.
-  ;; Don't maerge this with the pervious clause.
-  (if (string-match "values"
-		    (pp (byte-compile (lambda () (values t)))))
-      (defsubst values (&rest values)
-	values)))
+(require 'dabbrev)
+(require 'compile)
+(require 'flymake)
+(require 'outline)
+(require 'haskell-align-imports)
+(require 'haskell-sort-imports)
+(require 'haskell-string)
+(with-no-warnings (require 'cl))
+
+;; FIXME: code-smell: too many forward decls for haskell-session are required here
+(defvar haskell-session)
+(declare-function haskell-process-do-try-info "haskell-process" (sym))
+(declare-function haskell-process-generate-tags "haskell-process" (&optional and-then-find-this-tag))
+(declare-function haskell-session "haskell-session" ())
+(declare-function haskell-session-all-modules "haskell-session" (&optional DONTCREATE))
+(declare-function haskell-session-cabal-dir "haskell-session" (session))
+(declare-function haskell-session-maybe "haskell-session" ())
+(declare-function haskell-session-tags-filename "haskell-session" (session))
 
 ;; All functions/variables start with `(literate-)haskell-'.
 
 ;; Version of mode.
-(defconst haskell-version "v2.8.0"
-  "`haskell-mode' version number.")
-(defun haskell-version ()
-  "Echo the current version of `haskell-mode' in the minibuffer."
+(defconst haskell-version "@VERSION@"
+  "The release version of `haskell-mode'.")
+
+(defconst haskell-git-version "@GIT_VERSION@"
+  "The Git version of `haskell-mode'.")
+
+(defvar haskell-mode-pkg-base-dir (file-name-directory load-file-name)
+  "Package base directory of installed `haskell-mode'.
+Used for locating additional package data files.")
+
+;;;###autoload
+(defun haskell-version (&optional here)
+  "Show the `haskell-mode` version in the echo area.
+With prefix argument HERE, insert it at point.
+When FULL is non-nil, use a verbose version string.
+When MESSAGE is non-nil, display a message with the version."
+  (interactive "P")
+  (let* ((haskell-mode-dir (ignore-errors
+                             (file-name-directory (or (locate-library "haskell-mode") ""))))
+         (_version (format "haskell-mode version %s (%s @ %s)"
+                            haskell-version
+                            haskell-git-version
+                            haskell-mode-dir)))
+    (if here
+        (insert _version)
+      (message "%s" _version))))
+
+;;;###autoload
+(defun haskell-mode-view-news ()
+  "Display information on recent changes to haskell-mode."
   (interactive)
-  (message "Using haskell-mode version %s" haskell-version))
+  (with-current-buffer (find-file-read-only (expand-file-name "NEWS" haskell-mode-pkg-base-dir))
+    (goto-char (point-min))
+    (hide-sublevels 1)
+    (outline-next-visible-heading 1)
+    (show-subtree)))
 
 (defgroup haskell nil
   "Major mode for editing Haskell programs."
+  :link '(custom-manual "(haskell-mode)")
   :group 'languages
   :prefix "haskell-")
 
-;; Set load-path
 ;;;###autoload
-(add-to-list 'load-path
-   (or (file-name-directory load-file-name) (car load-path)))
-
-;; Set up autoloads for the modules we supply
-(autoload 'turn-on-haskell-decl-scan "haskell-decl-scan"
-  "Turn on Haskell declaration scanning." t)
-(autoload 'turn-on-haskell-doc-mode "haskell-doc"
-  "Turn on Haskell Doc minor mode." t)
-(autoload 'turn-on-haskell-indentation "haskell-indentation"
-  "Turn on advanced Haskell indentation." t)
-(autoload 'turn-on-haskell-indent "haskell-indent"
-  "Turn on Haskell indentation." t)
-(autoload 'turn-on-haskell-simple-indent "haskell-simple-indent"
-  "Turn on simple Haskell indentation." t)
-
-;; Functionality provided in other files.
-(autoload 'haskell-ds-create-imenu-index "haskell-decl-scan")
-(autoload 'haskell-font-lock-choose-keywords "haskell-font-lock")
-(autoload 'haskell-doc-current-info "haskell-doc")
-
-;; Obsolete functions.
-(defun turn-on-haskell-font-lock ()
-  (turn-on-font-lock)
-  (message "turn-on-haskell-font-lock is obsolete.  Use turn-on-font-lock instead."))
-(defun turn-on-haskell-hugs () (message "haskell-hugs is obsolete."))
-(defun turn-on-haskell-ghci () (message "haskell-ghci is obsolete."))
-
+(defun haskell-customize ()
+  "Browse the haskell customize sub-tree.
+This calls 'customize-browse' with haskell as argument and makes
+sure all haskell customize definitions have been loaded."
+  (interactive)
+  ;; make sure all modules with (defcustom ...)s are loaded
+  (mapc 'require
+        '(haskell-checkers haskell-doc haskell-font-lock haskell-indentation haskell-indent haskell-interactive-mode haskell-menu haskell-process haskell-yas inf-haskell))
+  (customize-browse 'haskell))
 
 ;; Are we looking at a literate script?
 (defvar haskell-literate nil
@@ -230,13 +219,15 @@ be set to the preferred literate style."
   :group 'haskell
   :type '(choice (const bird) (const tex) (const nil)))
 
-;; Mode maps.
+;;;###autoload
 (defvar haskell-mode-map
   (let ((map (make-sparse-keymap)))
-    ;; Bindings for the inferior haskell process:
+    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+    ;; For inferior haskell mode, use the below bindings
     ;; (define-key map [?\M-C-x]     'inferior-haskell-send-defun)
     ;; (define-key map [?\C-x ?\C-e] 'inferior-haskell-send-last-sexp)
     ;; (define-key map [?\C-c ?\C-r] 'inferior-haskell-send-region)
+    (define-key map [?\C-x ?\C-d] 'inferior-haskell-send-decl)
     (define-key map [?\C-c ?\C-z] 'switch-to-haskell)
     (define-key map [?\C-c ?\C-l] 'inferior-haskell-load-file)
     ;; I think it makes sense to bind inferior-haskell-load-and-run to C-c
@@ -247,15 +238,18 @@ be set to the preferred literate style."
     ;; (define-key map [?\C-c ?\C-s] 'inferior-haskell-start-process)
     ;; That's what M-; is for.
     ;; (define-key map "\C-c\C-c" 'comment-region)
-
     (define-key map (kbd "C-c C-t") 'inferior-haskell-type)
     (define-key map (kbd "C-c C-i") 'inferior-haskell-info)
     (define-key map (kbd "C-c M-.") 'inferior-haskell-find-definition)
     (define-key map (kbd "C-c C-d") 'inferior-haskell-find-haddock)
-
     (define-key map [?\C-c ?\C-v] 'haskell-check)
 
+    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+    ;; Editing-specific commands
+    (define-key map (kbd "C-c C-.") 'haskell-mode-format-imports)
     (define-key map [remap delete-indentation] 'haskell-delete-indentation)
+    (define-key map [backtab] 'unindent-for-tab-command)
+
     map)
   "Keymap used in Haskell mode.")
 
@@ -327,7 +321,7 @@ be set to the preferred literate style."
 	       (modify-syntax-entry i "_" table))
 	     (setq i (1+ i)))))
        (standard-syntax-table)))
-    
+
     (modify-syntax-entry ?\` "$`" table)
     (modify-syntax-entry ?\\ "\\" table)
     (mapc (lambda (x)
@@ -338,14 +332,14 @@ be set to the preferred literate style."
       ;; Non-ASCII syntax should be OK, at least in Emacs.
       (mapc (lambda (x)
               (modify-syntax-entry x "_" table))
-            (concat ""
-                    ""))
+            (concat "¡¢£¤¥¦§¨©ª«¬­®¯°±²³´µ¶·¸¹º»¼½¾¿"
+                    "×÷"))
       (mapc (lambda (x)
               (modify-syntax-entry x "w" table))
-            (concat ""
-                    ""
-                    ""
-                    "")))
+            (concat "ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖ"
+                    "ØÙÚÛÜÝÞß"
+                    "àáâãäåæçèéêëìíîïðñòóôõö"
+                    "øùúûüýþÿ")))
     table)
   "Syntax table used in Haskell mode.")
 
@@ -353,12 +347,18 @@ be set to the preferred literate style."
   "Return the identifier under point, or nil if none found.
 May return a qualified name."
   (save-excursion
+    ;; Skip whitespace if we're on it.  That way, if we're at "map ", we'll
+    ;; see the word "map".
+    (if (and (not (eobp))
+             (eq ?  (char-syntax (char-after))))
+        (skip-chars-backward " \t"))
+
     (let ((case-fold-search nil))
       (multiple-value-bind (start end)
           (if (looking-at "\\s_")
-              (values (progn (skip-syntax-backward "_") (point))
-                      (progn (skip-syntax-forward "_") (point)))
-            (values
+              (list (progn (skip-syntax-backward "_") (point))
+                    (progn (skip-syntax-forward "_") (point)))
+            (list
              (progn (skip-syntax-backward "w'")
                     (skip-syntax-forward "'") (point))
              (progn (skip-syntax-forward "w'") (point))))
@@ -395,25 +395,75 @@ May return a qualified name."
 
 (defcustom haskell-mode-hook nil
   "Hook run after entering Haskell mode.
-Do not select more than one of the three indentation modes."
+
+--------------------------------------------------------------------------------
+
+CONFIGURING INDENTATION
+
+  Using this you can configure the Haskell indentation mode. There
+  are three major Haskell indentation modes:
+
+  `haskell-indentation', Kristof Bastiaensen
+    Intelligent semi-automatic indentation, mark two. How to enable:
+    (custom-set-variables
+     '(haskell-mode-hook '(turn-on-haskell-indentation)))
+
+  `haskell-indent', Guy Lapalme
+    Intelligent semi-automatic indentation. How to enable:
+    (custom-set-variables
+     '(haskell-mode-hook '(turn-on-haskell-indentation)))
+
+  `haskell-simple-indent', Graeme E Moss and Heribert Schuetz
+    Simple indentation. How to enable:
+    (custom-set-variables
+     '(haskell-mode-hook '(turn-on-haskell-simple-indent)))
+
+  You can either:
+
+   1) Use the code above if you're more Elisp savvy, and put it
+      in your .emacs or similar file (type C-M-x to run each
+      one), or
+
+   2) customize the variable by ``M-x customize-group'' (see the
+      link below), or
+
+   3) some people prefer to add custom hooks like the below:
+
+      (add-hook 'haskell-mode-hook 'turn-on-haskell-indentation)
+
+  In order to test each one after enabling you can re-run M-x
+  haskell-mode in the same buffer.
+
+  Do not select more than one of the three indentation modes.
+
+--------------------------------------------------------------------------------"
   :type 'hook
   :group 'haskell
   :options `(turn-on-haskell-indent turn-on-haskell-indentation
-	     turn-on-font-lock
-	     ,(if (boundp 'eldoc-documentation-function)
-		  'turn-on-eldoc-mode
-		'turn-on-haskell-doc-mode) ; Emacs 21
-	     ,@(if (fboundp 'capitalized-words-mode)
-		   '(capitalized-words-mode))
-	     turn-on-simple-indent turn-on-haskell-doc-mode
-	     turn-on-haskell-decl-scan imenu-add-menubar-index))
+                                    turn-on-font-lock
+                                    ,(if (boundp 'eldoc-documentation-function)
+                                         'turn-on-eldoc-mode
+                                       'turn-on-haskell-doc-mode) ; Emacs 21
+                                    ,@(if (fboundp 'capitalized-words-mode)
+                                          '(capitalized-words-mode))
+                                    turn-on-haskell-simple-indent turn-on-haskell-doc-mode
+                                    turn-on-haskell-decl-scan imenu-add-menubar-index
+                                    turn-on-haskell-unicode-input-method))
 
 (defvar eldoc-print-current-symbol-info-function)
 
+;; For compatibility with Emacs < 24, derive conditionally
+(defalias 'haskell-parent-mode
+  (if (fboundp 'prog-mode) 'prog-mode 'fundamental-mode))
+
 ;; The main mode functions
 ;;;###autoload
-(define-derived-mode haskell-mode fundamental-mode "Haskell"
+(define-derived-mode haskell-mode haskell-parent-mode "Haskell"
   "Major mode for editing Haskell programs.
+
+See also Info node `(haskell-mode)Getting Started' for more
+information about this mode.
+
 Blank lines separate paragraphs, comments start with `-- '.
 \\<haskell-mode-map>
 Literate scripts are supported via `literate-haskell-mode'.
@@ -447,6 +497,7 @@ that `haskell-doc' is irregular in using `turn-(on/off)-haskell-doc-mode'.)
 Use `haskell-version' to find out what version this is.
 
 Invokes `haskell-mode-hook'."
+  :group 'haskell
   (set (make-local-variable 'paragraph-start) (concat "^$\\|" page-delimiter))
   (set (make-local-variable 'paragraph-separate) paragraph-start)
   (set (make-local-variable 'fill-paragraph-function) 'haskell-fill-paragraph)
@@ -458,6 +509,7 @@ Invokes `haskell-mode-hook'."
   (set (make-local-variable 'comment-end) "")
   (set (make-local-variable 'comment-end-skip) "[ \t]*\\(-}\\|\\s>\\)")
   (set (make-local-variable 'parse-sexp-ignore-comments) nil)
+  (set (make-local-variable 'indent-line-function) 'haskell-mode-suggest-indent-choice)
   ;; Set things up for eldoc-mode.
   (set (make-local-variable 'eldoc-documentation-function)
        'haskell-doc-current-info)
@@ -479,16 +531,49 @@ Invokes `haskell-mode-hook'."
   ;; TABs stops are 8 chars apart, as mandated by the Haskell Report.  --Stef
   (set (make-local-variable 'indent-tabs-mode) nil)
   (set (make-local-variable 'tab-width) 8)
-  (setq haskell-literate nil))
-
-(defun in-comment () (nth 4 (syntax-ppss)))
+  ;; dynamic abbrev support: recognize Haskell identifiers
+  ;; Haskell is case-sensitive language
+  (set (make-local-variable 'dabbrev-case-fold-search) nil)
+  (set (make-local-variable 'dabbrev-case-distinction) nil)
+  (set (make-local-variable 'dabbrev-case-replace) nil)
+  (set (make-local-variable 'dabbrev-abbrev-char-regexp) "\\sw\\|[.]")
+  (setq haskell-literate nil)
+  (add-hook 'before-save-hook 'haskell-mode-before-save-handler nil t)
+  (add-hook 'after-save-hook 'haskell-mode-after-save-handler nil t)
+  )
 
 (defun haskell-fill-paragraph (justify)
   (save-excursion
-    ;; We don't want to reflow code.
-    (unless (in-comment)
-      (end-of-line)) ; Try to get inside a comment
-    (if (in-comment) nil t)))
+    ;; Fill paragraph should only work in comments.
+    ;; The -- comments are handled properly by default
+    ;; The {- -} comments need some extra love.
+    (let* ((syntax-values (syntax-ppss))
+           (comment-num (nth 4 syntax-values)))
+      (cond
+       ((eq t comment-num)
+        ;; standard fill works wonders inside a non-nested comment
+        (fill-comment-paragraph justify))
+
+       ((integerp comment-num)
+        ;; we are in a nested comment. lets narrow to comment content
+        ;; and use plain paragraph fill for that
+        (let* ((comment-start-point (nth 8 syntax-values))
+               (comment-end-point
+                (save-excursion
+                  (re-search-forward "-}" (point-max) t comment-num)
+                  (point)))
+               (fill-paragraph-handle-comment nil))
+          (save-restriction
+            (narrow-to-region (+ 2 comment-start-point) (- comment-end-point 2))
+            (fill-paragraph justify))))
+       ((eolp)
+        ;; do nothing outside of a comment
+        t)
+       (t
+        ;; go to end of line and try again
+        (end-of-line)
+        (haskell-fill-paragraph justify))))))
+
 
 ;; (defun haskell-adaptive-fill ()
 ;;   ;; We want to use "--  " as the prefix of "-- |", etc.
@@ -502,7 +587,7 @@ Invokes `haskell-mode-hook'."
 ;;         (let ();(prefix-start (point)))
 ;;           (skip-syntax-forward "^w")
 ;;           (make-string (- (point) line-start) ?\s))))))
-          
+
 
 
 ;;;###autoload
@@ -548,13 +633,15 @@ If nil, use the Hoogle web-site."
                         nil nil def))))
   (if (null haskell-hoogle-command)
       (browse-url (format "http://haskell.org/hoogle/?q=%s" query))
-    (if (fboundp 'help-setup-xref)
-        (help-setup-xref (list 'haskell-hoogle query) (interactive-p)))
-    (with-output-to-temp-buffer
-        (if (fboundp 'help-buffer) (help-buffer) "*Help*")
-      (with-current-buffer standard-output
-        (start-process "hoogle" (current-buffer) haskell-hoogle-command
-                       query)))))
+    (lexical-let ((temp-buffer (if (fboundp 'help-buffer) (help-buffer) "*Help*")))
+      (with-output-to-temp-buffer temp-buffer
+	(with-current-buffer standard-output
+	  (let ((hoogle-process
+		 (start-process "hoogle" (current-buffer) haskell-hoogle-command query))
+		(scroll-to-top
+		 (lambda (process event)
+		   (set-window-start (get-buffer-window temp-buffer t) 1))))
+	    (set-process-sentinel hoogle-process scroll-to-top)))))))
 
 ;;;###autoload
 (defalias 'hoogle 'haskell-hoogle)
@@ -581,6 +668,16 @@ If nil, use the Hoogle web-site."
 		 (const "ghc -fno-code")
 		 (string :tag "Other command")))
 
+(defcustom haskell-stylish-on-save nil
+  "Whether to run stylish-haskell on the buffer before saving."
+  :group 'haskell
+  :type 'boolean)
+
+(defcustom haskell-tags-on-save nil
+  "Generate tags via hasktags after saving."
+  :group 'haskell
+  :type 'boolean)
+
 (defvar haskell-saved-check-command nil
   "Internal use.")
 
@@ -597,30 +694,198 @@ See `haskell-check-command' for the default."
 				    (if name
 					(file-name-nondirectory name))))))))
   (setq haskell-saved-check-command command)
-  (require 'compile)
   (save-some-buffers (not compilation-ask-about-save) nil)
-  (if (fboundp 'compilation-start)
-      (compilation-start command)
-    (compile-internal command "No more errors")))
-
-(autoload 'flymake-init-create-temp-buffer-copy "flymake")
+  (compilation-start command))
 
 (defun haskell-flymake-init ()
   "Flymake init function for Haskell.
 To be added to `flymake-init-create-temp-buffer-copy'."
-  (let ((checker-elts (split-string haskell-saved-check-command)))
+  (let ((checker-elts (and haskell-saved-check-command
+			   (split-string haskell-saved-check-command))))
     (list (car checker-elts)
 	  (append (cdr checker-elts)
 		  (list (flymake-init-create-temp-buffer-copy
 			 'flymake-create-temp-inplace))))))
 
-(eval-after-load "flymake"
-  '(add-to-list 'flymake-allowed-file-name-masks
-		'("\\.l?hs\\'" haskell-flymake-init)))
+(add-to-list 'flymake-allowed-file-name-masks '("\\.l?hs\\'" haskell-flymake-init))
+
+(defun haskell-mode-suggest-indent-choice ()
+  "Ran when the user tries to indent in the buffer but no indentation mode has been selected.
+Brings up the documentation for haskell-mode-hook."
+  (describe-variable 'haskell-mode-hook))
+
+(defvar unindent-line-function nil
+  "Function to unindent the current line.
+This function will be called with no arguments.")
+
+(defun unindent-for-tab-command ()
+  "Un-indent the current line according to the mode's unindenting function (if any)."
+  (interactive)
+  (when unindent-line-function
+    (funcall unindent-line-function)))
+
+(defun haskell-mode-format-imports ()
+  "Format the imports by aligning and sorting them."
+  (interactive)
+  (let ((col (current-column)))
+    (haskell-sort-imports)
+    (haskell-align-imports)
+    (goto-char (+ (line-beginning-position)
+                  col))))
+
+(defun haskell-mode-message-line (str)
+  "Message only one line, multiple lines just disturbs the programmer."
+  (let ((lines (split-string str "\n" t)))
+    (when (and (car lines) (stringp (car lines)))
+      (message "%s"
+               (concat (car lines)
+                       (if (and (cdr lines) (stringp (cadr lines)))
+                           (format " [ %s .. ]" (haskell-string-take (haskell-trim (cadr lines)) 10))
+                         ""))))))
+
+(defun haskell-mode-contextual-space ()
+  "Contextually do clever stuff when hitting space."
+  (interactive)
+  (if (not (haskell-session-maybe))
+      (self-insert-command 1)
+    (cond ((save-excursion (forward-word -1)
+                           (looking-at "^import$"))
+           (insert " ")
+           (let ((module (ido-completing-read "Module: " (haskell-session-all-modules))))
+             (insert module)
+             (haskell-mode-format-imports)))
+          ((not (string= "" (save-excursion (forward-char -1) (haskell-ident-at-point))))
+           (let ((ident (save-excursion (forward-char -1) (haskell-ident-at-point))))
+             (insert " ")
+             (haskell-process-do-try-info ident)))
+          (t (insert " ")))))
+
+(defun haskell-mode-before-save-handler ()
+  "Function that will be called before buffer's saving."
+  )
+
+(defun haskell-mode-after-save-handler ()
+  "Function that will be called after buffer's saving."
+  (when haskell-tags-on-save
+    (ignore-errors (when (and (boundp 'haskell-session) haskell-session)
+                     (haskell-process-generate-tags))))
+  (when haskell-stylish-on-save
+    (ignore-errors (haskell-mode-stylish-buffer)))
+  (let ((before-save-hook '())
+        (after-save-hook '()))
+    (basic-save-buffer))
+  )
+
+(defun haskell-mode-buffer-apply-command (cmd)
+  "Execute shell command CMD with current buffer as input and
+replace the whole buffer with the output. If CMD fails the buffer
+remains unchanged."
+  (set-buffer-modified-p t)
+  (let* ((chomp (lambda (str)
+                  (while (string-match "\\`\n+\\|^\\s-+\\|\\s-+$\\|\n+\\'" str)
+                    (setq str (replace-match "" t t str)))
+                  str))
+         (errout (lambda (fmt &rest args)
+                   (let* ((warning-fill-prefix "    "))
+                     (display-warning cmd (apply 'format fmt args) :warning))))
+         (filename (buffer-file-name (current-buffer)))
+         (cmd-prefix (replace-regexp-in-string " .*" "" cmd))
+         (tmp-file (make-temp-file cmd-prefix))
+         (err-file (make-temp-file cmd-prefix))
+         (default-directory (if (and (boundp 'haskell-session)
+                                     haskell-session)
+                                (haskell-session-cabal-dir haskell-session)
+                              default-directory))
+         (errcode (with-temp-file tmp-file
+                    (call-process cmd filename
+                                  (list (current-buffer) err-file) nil)))
+         (stderr-output
+          (with-temp-buffer
+            (insert-file-contents err-file)
+            (funcall chomp (buffer-substring-no-properties (point-min) (point-max)))))
+         (stdout-output
+          (with-temp-buffer
+            (insert-file-contents tmp-file)
+            (buffer-substring-no-properties (point-min) (point-max)))))
+      (if (string= "" stderr-output)
+          (if (string= "" stdout-output)
+              (funcall errout
+               "Error: %s produced no output, leaving buffer alone" cmd)
+            (save-restriction
+              (widen)
+              ;; command successful, insert file with replacement to preserve
+              ;; markers.
+              (insert-file-contents tmp-file nil nil nil t)))
+        ;; non-null stderr, command must have failed
+        (funcall errout "%s failed: %s" cmd stderr-output)
+        )
+      (delete-file tmp-file)
+      (delete-file err-file)
+      ))
+
+(defun haskell-mode-stylish-buffer ()
+  "Apply stylish-haskell to the current buffer."
+  (interactive)
+  (let ((column (current-column))
+        (line (line-number-at-pos)))
+    (haskell-mode-buffer-apply-command "stylish-haskell")
+    (goto-char (point-min))
+    (forward-line (1- line))
+    (goto-char (+ column (point)))))
+
+(defun haskell-mode-tag-find (&optional next-p)
+  "The tag find function, specific for the particular session."
+  (interactive "P")
+  (let ((tags-file-name (haskell-session-tags-filename (haskell-session)))
+        (tags-revert-without-query t)
+        (ident (haskell-ident-at-point)))
+    (when (not (string= "" (haskell-trim ident)))
+      (cond ((file-exists-p tags-file-name)
+             (find-tag ident next-p))
+            (t (haskell-process-generate-tags ident))))))
+
+;; From Bryan O'Sullivan's blog:
+;; http://www.serpentine.com/blog/2007/10/09/using-emacs-to-insert-scc-annotations-in-haskell-code/
+(defun haskell-mode-insert-scc-at-point ()
+  "Insert an SCC annotation at point."
+  (interactive)
+  (if (or (looking-at "\\b\\|[ \t]\\|$") (and (not (bolp))
+					  (save-excursion
+					    (forward-char -1)
+					    (looking-at "\\b\\|[ \t]"))))
+      (let ((space-at-point (looking-at "[ \t]")))
+	(unless (and (not (bolp)) (save-excursion
+				    (forward-char -1)
+				    (looking-at "[ \t]")))
+	  (insert " "))
+	(insert "{-# SCC \"\" #-}")
+	(unless space-at-point
+	  (insert " "))
+	(forward-char (if space-at-point -5 -6)))
+    (error "Not over an area of whitespace")))
+
+;; Also Bryan O'Sullivan's.
+(defun haskell-mode-kill-scc-at-point ()
+  "Kill the SCC annotation at point."
+  (interactive)
+  (save-excursion
+    (let ((old-point (point))
+	  (scc "\\({-#[ \t]*SCC \"[^\"]*\"[ \t]*#-}\\)[ \t]*"))
+      (while (not (or (looking-at scc) (bolp)))
+	(forward-char -1))
+      (if (and (looking-at scc)
+	       (<= (match-beginning 1) old-point)
+	       (> (match-end 1) old-point))
+	  (kill-region (match-beginning 0) (match-end 0))
+	(error "No SCC at point")))))
+
 
 ;; Provide ourselves:
 
 (provide 'haskell-mode)
 
-;; arch-tag: b2237ec0-ddb0-4c86-9339-52d410264980
+;; Local Variables:
+;; byte-compile-warnings: (not cl-functions)
+;; End:
+
 ;;; haskell-mode.el ends here
