@@ -13,7 +13,7 @@
 
 ;; This file is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
-;; the Free Software Foundation; either version 2, or (at your option)
+;; the Free Software Foundation; either version 3, or (at your option)
 ;; any later version.
 
 ;; This file is distributed in the hope that it will be useful,
@@ -22,11 +22,8 @@
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs; see the file COPYING.  If not, write to the
-;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-;; Boston, MA 02111-1307, USA.
+;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-
 ;;; Commentary:
 
 ;; Purpose:
@@ -68,6 +65,9 @@
 ;; All functions/variables start with
 ;; `(turn-(on/off)-)haskell-simple-indent'.
 
+;; Dynamically scoped variables.
+(defvar unindent-line-function)
+
 ;; Version.
 (defconst haskell-simple-indent-version "1.2"
   "`haskell-simple-indent' version number.")
@@ -123,9 +123,49 @@ column, `tab-to-tab-stop' is done instead."
 	  (set-marker opoint nil))
       (tab-to-tab-stop))))
 
+(defun haskell-simple-indent-backtab ()
+  "Indent backwards."
+  (interactive)
+  (let ((current-point (point))
+        (i 0)
+        (x 0))
+    (goto-char (line-beginning-position))
+    (save-excursion
+      (while (< (point) current-point)
+        (haskell-simple-indent)
+        (setq i (+ i 1))))
+    (while (< x (- i 1))
+      (haskell-simple-indent)
+      (setq x (+ x 1)))))
+
+(defun haskell-simple-indent-newline-same-col ()
+  "Make a newline and go to the same column as the current line."
+  (interactive)
+  (let ((point (point)))
+    (let ((start-end
+	   (save-excursion
+	     (let* ((start (line-beginning-position))
+		    (end (progn (goto-char start)
+				(search-forward-regexp
+				 "[^ ]" (line-end-position) t 1))))
+	       (when end (cons start (1- end)))))))
+      (if start-end
+	  (progn (newline)
+		 (insert (buffer-substring-no-properties
+			  (car start-end) (cdr start-end))))
+	(newline)))))
+
+(defun haskell-simple-indent-newline-indent ()
+  "Make a newline on the current column and indent on step."
+  (interactive)
+  (haskell-simple-indent-newline-same-col)
+  (insert "  "))
+
 (defvar haskell-simple-indent-old)
+(defvar haskell-simple-unindent-old)
 
 ;; The main functions.
+;;;###autoload
 (defun turn-on-haskell-simple-indent ()
   "Set `indent-line-function' to a simple indentation function.
 TAB will now move the cursor to the next indent point in the previous
@@ -135,20 +175,25 @@ whitespace.
 Runs `haskell-simple-indent-hook'.
 
 Use `haskell-simple-indent-version' to find out what version this is."
+  (interactive)
   (set (make-local-variable 'haskell-simple-indent-old) indent-line-function)
   (set (make-local-variable 'indent-line-function) 'haskell-simple-indent)
+  (set (make-local-variable 'haskell-simple-unindent-old) unindent-line-function)
+  (set (make-local-variable 'unindent-line-function) 'haskell-simple-indent-backtab)
   (run-hooks 'haskell-simple-indent-hook))
 
 (defun turn-off-haskell-simple-indent ()
   "Return `indent-line-function' to original value.
 I.e. the value before `turn-on-haskell-simple-indent' was called."
+  (interactive)
   (when (local-variable-p 'haskell-simple-indent-old)
     (setq indent-line-function haskell-simple-indent-old)
-    (kill-local-variable 'haskell-simple-indent-old)))
+    (setq unindent-line-function haskell-simple-unindent-old)
+    (kill-local-variable 'haskell-simple-indent-old)
+    (kill-local-variable 'haskell-simple-unindent-old)))
 
 ;; Provide ourselves:
 
 (provide 'haskell-simple-indent)
 
-;; arch-tag: 18a08122-723b-485e-b958-e1cf8218b816
 ;;; haskell-simple-indent.el ends here
